package dk.nsp.idsas.integrations.sores;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import dk.nsp.idsas.integrations.ServiceErrorCollector;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.entity.StringEntity;
import org.apache.log4j.Logger;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.stream.Collectors;

public class SoresClient {
    private final static Logger logger = Logger.getLogger(SoresClient.class);

    public static String ERROR_MSG = "Fejl i SORES-integrationen";
    public static String ERROR_MSG_BAD_REQUEST = "SORES svarede med status 400";
    public static String ERROR_MSG_INTERNAL_SERVER_ERROR = "SORES svarede med status 500";

    private static final String CONTENT_TYPE = "application/json";

    private final ObjectMapper objectMapper = new ObjectMapper();
    private final String endpoint;
    private final HttpClient httpClient;
    private final ServiceErrorCollector serviceErrorCollector;

    public SoresClient(HttpClient httpClient, String endpoint, ServiceErrorCollector serviceErrorCollector) {
        this.endpoint = endpoint;
        this.httpClient = httpClient;
        this.serviceErrorCollector = serviceErrorCollector;
    }

    public int getErrorCount() {
        return serviceErrorCollector.getErrorCount();
    }

    public Map<String,String> getCvrNumberBySor(String[] sor) throws SoresClientException {
        // Create new array where non-integer Strings are removed:
        String[] validatedInputSor = Arrays.stream(sor)
                .filter(s -> s.matches("\\d+"))
                .toArray(String[]::new);

        try {
            String body = buildInstitutionOwnerCvrNumberBody(validatedInputSor);
            HttpUriRequest request = buildSorRequest(body);
            HttpResponse response = invokeSor(request);
            Map<String,String> result = handleResponse(validatedInputSor, response);
            serviceErrorCollector.registerServiceSuccess();
            return result;
        } catch (UnsupportedOperationException | IOException e) {
            serviceErrorCollector.registerServiceError();
            throw new SoresClientException(ERROR_MSG, e);
        }
    }

    public boolean checkSorExists(String sorCode) throws SoresClientException {
        try {
            HttpUriRequest request = buildSorExistsRequest(sorCode);
            HttpResponse response = invokeSor(request);
            String[] sorCodeArray = { sorCode };
            Map<String,String> result = handleResponse(sorCodeArray, response);
            serviceErrorCollector.registerServiceSuccess();

            JsonNode jsonNode = objectMapper.readTree(result.get(sorCode));
            return !jsonNode.isEmpty();

        } catch (UnsupportedOperationException | IOException e) {
            serviceErrorCollector.registerServiceError();
            throw new SoresClientException(ERROR_MSG, e);
        }
    }

    private HttpUriRequest buildSorRequest(String input) throws UnsupportedEncodingException {
        final StringEntity entity = new StringEntity(input);
        final HttpPost httpPost = new HttpPost(endpoint);
        httpPost.setEntity(entity);
        httpPost.setHeader("Content-type", CONTENT_TYPE);
        return httpPost;
    }

    private String buildInstitutionOwnerCvrNumberBody(String[] inputSor) throws JsonProcessingException {
        final ObjectMapper objectMapper = new ObjectMapper();

        return objectMapper.writeValueAsString(new Object() {
            public final String op = "InstitutionOwnerCvrNumber";
            public final String[] sorIds = Arrays.copyOf(inputSor, inputSor.length);
        });
    }

    private HttpUriRequest buildSorExistsRequest(String sorCode) throws JsonProcessingException, UnsupportedEncodingException {
        final ObjectMapper objectMapper = new ObjectMapper();
        final String jsonString = objectMapper.writeValueAsString(new Object() {
            public final String op = "GetSorEntity";
            public final String[] sorIds = { sorCode };
        });
        final StringEntity entity = new StringEntity(jsonString);

        final HttpPost httpPost = new HttpPost(endpoint);
        httpPost.setEntity(entity);
        httpPost.setHeader("Content-type", CONTENT_TYPE);
        return httpPost;
    }

    private HttpResponse invokeSor(HttpUriRequest request) throws SoresClientException {
        try {
            return httpClient.execute(request);
        } catch (IOException e) {
            serviceErrorCollector.registerServiceError();
            throw new SoresClientException(ERROR_MSG, e);
        }
    }

    private Map<String,String> handleResponse(String[] inputCodes, HttpResponse response)
            throws SoresClientException, IOException {

        if (response.getStatusLine().getStatusCode() == HttpStatus.SC_INTERNAL_SERVER_ERROR) {
            serviceErrorCollector.registerServiceError();
            throw new SoresClientException(ERROR_MSG_INTERNAL_SERVER_ERROR);
        }

        if (response.getStatusLine().getStatusCode() == HttpStatus.SC_BAD_REQUEST) {
            serviceErrorCollector.registerServiceError();
            throw new SoresClientException(ERROR_MSG_BAD_REQUEST);
        }

        JsonNode jsonNode = objectMapper.readTree(response.getEntity().getContent());
        if (jsonNode.size() == 0) {
            logger.warn("SORES returned empty response");
            return new HashMap<>();
        } else {
            return Arrays.stream(inputCodes).collect(Collectors.toMap(
                    code -> code,
                    code -> jsonNode.get(code).isTextual() ?
                            jsonNode.get(code).asText() : jsonNode.get(code).toString()));
        }
    }

}
