package dk.nsp.idsas.service;

import dk.nsp.idsas.dao.OrgBlurringDao;
import dk.nsp.idsas.dao.WhitelistDao;
import dk.nsp.idsas.dao.entity.OrgBlurringEntity;
import dk.nsp.idsas.dto.OrgIdTypeDto;
import dk.nsp.idsas.dto.OrganizationIdDto;
import dk.nsp.idsas.integrations.sores.SoresClient;
import dk.nsp.idsas.integrations.sores.SoresClientException;
import dk.nsp.idsas.util.ErrorMessages;
import dk.nsp.idsas.util.FaultCode;
import dk.nsp.idsas.util.WhitelistType;
import org.apache.log4j.Logger;
import org.springframework.dao.DataAccessException;

import java.time.Instant;
import java.time.OffsetDateTime;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

public class OrgBlurringService {
    private static final Logger logger = Logger.getLogger(OrgBlurringService.class);

    private final WhitelistDao whitelistDao;
    private final OrgBlurringDao orgBlurringDao;
    private final SoresClient soresClient;
    private final String[] allowedClientSystemNames;

    public OrgBlurringService(WhitelistDao whitelistDao, OrgBlurringDao orgBlurringDao, SoresClient soresClient,
                              String[] allowedClientSystemNames) {
        this.whitelistDao = whitelistDao;
        this.orgBlurringDao = orgBlurringDao;
        this.soresClient = soresClient;
        this.allowedClientSystemNames = allowedClientSystemNames;
    }

    /**
     * Opretter en ny afdelingssløring.<br/>
     * Understøttede organisationer i createOrgBlurring er SOR og SHAK. SOR-koder valideres på om de eksisterer via kald til SORES.
     * SHAK koder kontrolleres ikke.<br/>
     */
    public void createOrgBlurring(String creatorCVR, String organizationId, String organizationIdType) throws IdsasServiceException {
        logger.trace("Create org blurring");

        // Tjek at creatorCVR er whitelisted til at kalde createBlurring / createOrgBlurring
        boolean isWhiteListed = whitelistDao.isWhitelisted(creatorCVR, WhitelistType.BLURRING);
        if (!isWhiteListed) {
            throw new IdsasServiceException(ErrorMessages.cvrWhitelistError(creatorCVR, WhitelistType.BLURRING),
                    FaultCode.CREATE_BLURRING_ERROR, true);
        }

        // Hvis der registreres med SOR-kode, skal vi tjekke at SOR-koden hører under opretterens CVR-nummer
        if (organizationIdType.equals("SOR")) {
            try {
                Map<String,String> sorToCvr = soresClient.getCvrNumberBySor(new String[]{organizationId});
                if (sorToCvr.get(organizationId) == null || !sorToCvr.get(organizationId).equals(creatorCVR)) {
                    throw new IdsasServiceException(ErrorMessages.sorDoesNotBelongToCvr(organizationId, creatorCVR), FaultCode.CREATE_ORG_BLURRING_ERROR, false);
                }
            } catch (SoresClientException e) {
                throw new IdsasServiceException(ErrorMessages.sorClientError(), FaultCode.CREATE_ORG_BLURRING_ERROR, false, e);
            }
        }

        OrgBlurringEntity orgBlurringEntity = new OrgBlurringEntity();
        orgBlurringEntity.setCreatedDate(OffsetDateTime.now());
        orgBlurringEntity.setCreatorCvr(creatorCVR);
        orgBlurringEntity.setOrganisationId(organizationId);
        orgBlurringEntity.setOrganisationIdType(organizationIdType);
        orgBlurringEntity.setValidFrom(Instant.now());
        orgBlurringEntity.setValidTo(null); // Ingen udløbsdato

        // Der må ikke allerede være en registrering
        try {
            if (orgBlurringDao.orgBlurringExists(organizationId, organizationIdType)) {
                throw new IdsasServiceException(ErrorMessages.orgBlurringAlreadyExists(), FaultCode.CREATE_ORG_BLURRING_ERROR, false);
            }
        } catch (DataAccessException e) {
            throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.CREATE_ORG_BLURRING_ERROR, false, e);
        }

        // SOR-koder skal eksistere
        if (organizationIdType.equals("SOR")) {
            try {
                if (!soresClient.checkSorExists(organizationId)) {
                    throw new IdsasServiceException(ErrorMessages.sorDoesNotExist(), FaultCode.CREATE_ORG_BLURRING_ERROR, false);
                }
            } catch (SoresClientException e) {
                throw new IdsasServiceException(ErrorMessages.sorClientError(), FaultCode.CREATE_ORG_BLURRING_ERROR, false, e);
            }
        }

        // Opret ny registering
        try {
            orgBlurringDao.insert(orgBlurringEntity);
        } catch (DataAccessException e) {
            throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.CREATE_ORG_BLURRING_ERROR, false, e);
        }
    }

    /**
     * Returnerer listen af registrerede afdelinger hørende til CVR nummeret i SOSI ID-kortet.
     */
    public List<OrganizationIdDto> listOrgBlurringsForCVR(String creatorCvr) throws IdsasServiceException {
        logger.trace("List org blurrings for CVR");
        try {
            List<OrgBlurringEntity> entities = orgBlurringDao.getAllOrgBlurrings(creatorCvr);
            return entities.stream()
                    .map(entity -> new OrganizationIdDto(
                            entity.getOrganisationId(),
                            OrgIdTypeDto.valueOf(entity.getOrganisationIdType())))
                    .collect(Collectors.toList());
        } catch (DataAccessException e) {
            throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.LIST_ORG_BLURRING_ERROR, false, e);
        }
    }

    /**
     * Returnerer listen af alle aktive afdelingssløringer. Skal bruges af tjenester, der skal effektuere afdelingssløringer.
     */
    public List<OrganizationIdDto> listAllActiveOrgBlurrings(String clientSystemName) throws IdsasServiceException {
        logger.trace("List all active org blurrings");

        // Validering af, at systembrugeren har et clientName med den korrekte værdi
        if (!Arrays.asList(allowedClientSystemNames).contains(clientSystemName)) {
            String names = String.join(", ", allowedClientSystemNames);
            throw new IdsasServiceException(ErrorMessages.clientSystemNameError(
                    clientSystemName, names), FaultCode.LOOKUP_ERROR, true);
        }

        try {
            List<OrgBlurringEntity> entities = orgBlurringDao.getAllActiveOrgBlurrings();
            return entities.stream()
                    .map(entity -> new OrganizationIdDto(
                            entity.getOrganisationId(),
                            OrgIdTypeDto.valueOf(entity.getOrganisationIdType())))
                    .collect(Collectors.toList());
        } catch (DataAccessException e) {
            throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.LIST_ORG_BLURRING_ERROR, false, e);
        }
    }

    /**
     * Inaktiverer en afdelingssløring per 'nu' (sætter validTo til 'nu').<br/>
     * Funktionen fejler hvis man forsøger at fjerne en ikke eksisterende afdelingssløring eller hvis man forsøger at fjerne en afdelingssløring, der ikke hører til kalderens CVR-nummer.
     */
    public String removeOrgBlurring(String creatorCvr, String organizationId, String organizationIdType) throws IdsasServiceException {
        logger.trace("Remove org blurring");

        // Hent den aktive slørring
        List<OrgBlurringEntity> orgBlurringEntities = orgBlurringDao.getActiveOrgBlurrings(organizationId, organizationIdType);
        if (orgBlurringEntities.isEmpty()) {
            throw new IdsasServiceException(ErrorMessages.orgBlurringDoesNotExist(), FaultCode.REMOVE_ORG_BLURRING_ERROR, false);
        }

        // Tjek at creatorCvr matcher på registreringen
        if (!orgBlurringEntities.get(0).getCreatorCvr().equals(creatorCvr)) {
            throw new IdsasServiceException(ErrorMessages.orgBlurringNotCreatedByOrg(), FaultCode.REMOVE_ORG_BLURRING_ERROR, false);
        }

        // Sløringen annulleres
        try {
            orgBlurringDao.removeOrgBlurring(orgBlurringEntities.get(0).getOrganisationId(), orgBlurringEntities.get(0).getOrganisationIdType());
        } catch (DataAccessException e) {
            throw new IdsasServiceException(ErrorMessages.daoAccessError(), FaultCode.REMOVE_ORG_BLURRING_ERROR, false, e);
        }

        return "OK";
    }

}
